"""Room environment compatible with gym.

This env uses the RoomDes (room_env/envs/des.py), and Memory classes.
This is a more generalized version than RoomEnv0.
"""
import logging
import os
import random
from copy import deepcopy
from typing import Tuple

import gym

from ..des import RoomDes
from ..memory import EpisodicMemory, SemanticMemory, ShortMemory
from ..policy import answer_question, encode_observation, manage_memory
from ..utils import seed_everything

logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class RoomEnv1(gym.Env):
    """The Room environment version 1.

    This env includes three state-action spaces. You have to choose which one of the
    three will be RL trained.

    Memory management.
        State: episodic, semantic, and short-term memory systems at time t
        Action: (0) Move the oldest short-term memory to the episodic,
                (1) to the semantic, or (2) forget it

    Question-answer
        State: episodic and semantic memory systems at time t
        Action: (0) Select the episodic memory system to answer the question, or
                (1) the semantic

    Encoding an observation to a short-term memory. The state space is
        (i) triple-based, (ii) text-based, or (iii) image-based.
        Triple
            State: [(head_i, relation_i, tail_i) | i is from 1 to N]
            Action: Choose one of the N triples (actions) to be encoded as
                    a short-term memory.
        Text
            State: [token_1, token_2, …, token_N]
            Action: This is actually now N^3, where the first, second and third are to
                    choose head, relation, and tail, respectively.
        Image
            State: An image with objects
            Action: Not sure yet …

    """

    metadata = {"render.modes": ["console"]}

    def __init__(
        self,
        des_size: str = "l",
        seed: int = 42,
        policies: dict = {
            "memory_management": "RL",
            "question_answer": "episodic_semantic",
            "encoding": "argmax",
        },
        capacity: dict = {"episodic": 16, "semantic": 16, "short": 1},
        question_prob: int = 1.0,
        observation_params: str = "perfect",
        allow_random_human: bool = False,
        allow_random_question: bool = False,
        total_episode_rewards: int = 128,
        pretrain_semantic: bool = False,
        check_resources: bool = True,
        varying_rewards: bool = False,
    ) -> None:
        """

        Args
        ----
        des_size: "xxs", "xs", "s", "m", or "l".
        seed: random seed number
        policies:
            memory_management:
                "RL": Reinforcement learning to learn the policy.
                "episodic": Always take action 1: move to the episodic.
                "semantic": Always take action 2: move to the semantic.
                "forget": Always take action 3: forget the oldest short-term memory.
                "random": Take one of the three actions uniform-randomly.
                "neural": Neural network policy
            question_answer:
                "RL": Reinforcement learning to learn the policy.
                "episodic_semantic": First look up the episodic and then the semantic.
                "semantic_episodic": First look up the semantic and then the episodic.
                "episodic": Only look up the episodic.
                "semantic": Only look up the semantic.
                "random": Take one of the two actions uniform-randomly.
                "neural": Neural network policy
            encoding:
                "RL": Reinforcement learning to learn the policy.
                "argmax": Take the triple with the highest score.
                "neural": Neural network policy
        capacity: memory capactiy of the agent.
            e.g., {"episodic": 1, "semantic": 1}
        question_prob: The probability of a question being asked at every observation.
        observation_params: At the moment this is only "perfect".
        allow_random_human: whether or not to generate a random human sequence.
        allow_random_question: whether or not to geneate a random question sequence.
        total_episode_rewards: total episode rewards
        pretrain_semantic: whether to prepopulate the semantic memory with ConceptNet
                           or not
        check_resources: whether to check the resources in the DES.
        varying_rewards: If true, then the rewards are scaled in every episode so that
             total_episode_rewards is total_episode_rewards.

        """
        self.seed = seed
        seed_everything(self.seed)
        self.policies = policies
        assert len([pol for pol in self.policies.values() if pol.lower() == "rl"]) == 1
        self.capacity = capacity
        self.question_prob = question_prob

        self.observation_params = observation_params

        self.allow_random_human = allow_random_human
        self.allow_random_question = allow_random_question
        self.total_episode_rewards = total_episode_rewards
        self.pretrain_semantic = pretrain_semantic
        self.check_resources = check_resources
        self.varying_rewards = varying_rewards

        # Our state space is quite complex. Here we just make a dummy observation space.
        # to bypass the sanity check.
        self.observation_space = gym.spaces.Discrete(1)

        if self.policies["memory_management"].lower() == "rl":
            # 0 for episodic, 1 for semantic, and 2 to forget
            self.action_space = gym.spaces.Discrete(3)
        if self.policies["question_answer"].lower() == "rl":
            # 0 for episodic and 1 for semantic
            self.action_space = gym.spaces.Discrete(2)
        if self.policies["encoding"].lower() == "rl":
            raise NotImplementedError

        self.des_size = des_size
        self.des = RoomDes(
            des_size=self.des_size,
            check_resources=self.check_resources,
        )
        assert 0 < self.question_prob <= 1

        self.init_memory_systems()

    def init_memory_systems(self) -> None:
        """Initialize the agent's memory systems."""
        self.memory_systems = {
            "episodic": EpisodicMemory(capacity=self.capacity["episodic"]),
            "semantic": SemanticMemory(capacity=self.capacity["semantic"]),
            "short": ShortMemory(capacity=self.capacity["short"]),
        }

        if self.pretrain_semantic:
            assert self.capacity["semantic"] > 0
            _ = self.memory_systems["semantic"].pretrain_semantic(
                self.des.semantic_knowledge,
                return_remaining_space=False,
                freeze=False,
            )

    def generate_sequences(self) -> None:
        """Generate human and question sequences in advance."""
        if self.observation_params.lower() == "perfect":
            if self.allow_random_human:
                self.human_sequence = random.choices(
                    list(self.des.humans), k=self.des.until + 1
                )
            else:
                self.human_sequence = (
                    self.des.humans * (self.des.until // len(self.des.humans) + 1)
                )[: self.des.until + 1]
        else:
            raise NotImplementedError

        if self.allow_random_question:
            self.question_sequence = [
                random.choice(self.human_sequence[: i + 1])
                for i in range(len(self.human_sequence))
            ]
        else:
            self.question_sequence = [self.human_sequence[0]]
            self.des.run()
            assert (
                len(self.des.states)
                == len(self.des.events) + 1
                == len(self.human_sequence)
            )
            for i in range(len(self.human_sequence) - 1):
                start = max(i + 2 - len(self.des.humans), 0)
                end = i + 2
                humans_observed = self.human_sequence[start:end]

                current_state = self.des.states[end - 1]
                humans_not_changed = []
                for j, human in enumerate(humans_observed):
                    observed_state = self.des.states[start + j]

                    is_changed = False
                    for to_check in ["object", "object_location"]:
                        if (
                            current_state[human][to_check]
                            != observed_state[human][to_check]
                        ):
                            is_changed = True
                    if not is_changed:
                        humans_not_changed.append(human)

                self.question_sequence.append(random.choice(humans_not_changed))

            self.des._initialize()

        effective_question_sequence = []
        for i, question in enumerate(self.question_sequence[:-1]):
            if random.random() < self.question_prob:
                effective_question_sequence.append(question)
            else:
                effective_question_sequence.append(None)
        # The last observation shouldn't have a question
        effective_question_sequence.append(None)
        self.question_sequence = effective_question_sequence

        assert len(self.human_sequence) == len(self.question_sequence)

        self.num_questions = sum(
            [True for question in self.question_sequence if question is not None]
        )
        if self.varying_rewards:
            self.CORRECT = self.total_episode_rewards / self.num_questions
            self.WRONG = -self.CORRECT
        else:
            self.CORRECT = 1
            self.WRONG = -1

    @staticmethod
    def extract_memory_entires(memory_systems: dict) -> dict:
        """Extract the entries from the Memory objects.
        Ars
        ---
        memory_systems: {"episodic": EpisodicMemory, "semantic": SemanticMemory,
                        "short": ShortMemory}

        Returns
        -------
        memory_systems_: memory_systems only with entries.
        """
        memory_systems_ = {}
        for key, value in memory_systems.items():
            memory_systems_[key] = deepcopy(value.entries)

        return memory_systems_

    def generate_oqa(
        self, increment_des: bool = False
    ) -> Tuple[dict, dict, dict, bool]:
        """Generate an observation, question, and answer.

        Args
        ----
        increment_des: whether or not to take a step in the DES.

        Returns
        -------
        observation = {
            "human": <human>,
            "object": <obj>,
            "object_location": <obj_loc>,
        }
        question = {"human": <human>, "object": <obj>}
        answer = <obj_loc>
        is_last: True, if its the last observation in the queue, othewise False

        """
        human_o = self.human_sequence.pop(0)
        human_q = self.question_sequence.pop(0)

        is_last_o = len(self.human_sequence) == 0
        is_last_q = len(self.question_sequence) == 0

        assert is_last_o == is_last_q
        is_last = is_last_o

        if increment_des:
            self.des.step()

        obj_o = self.des.state[human_o]["object"]
        obj_loc_o = self.des.state[human_o]["object_location"]
        observation = deepcopy(
            {
                "human": human_o,
                "object": obj_o,
                "object_location": obj_loc_o,
                "current_time": self.des.current_time,
            }
        )

        if human_q is not None:
            obj_q = self.des.state[human_q]["object"]
            obj_loc_q = self.des.state[human_q]["object_location"]

            question = deepcopy({"human": human_q, "object": obj_q})
            answer = deepcopy(obj_loc_q)

        else:
            question = None
            answer = None

        return observation, question, answer, is_last

    def reset(self) -> dict:
        """Reset the environment.


        Returns
        -------
        state

        """
        self.des._initialize()
        self.generate_sequences()
        self.init_memory_systems()
        info = {}
        self.obs, self.question, self.answer, self.is_last = self.generate_oqa(
            increment_des=False
        )

        if self.policies["encoding"].lower() == "rl":
            return deepcopy(self.obs), info

        if self.policies["memory_management"].lower() == "rl":
            encode_observation(self.memory_systems, self.policies["encoding"], self.obs)
            return deepcopy(self.extract_memory_entires(self.memory_systems)), info

        if self.policies["question_answer"].lower() == "rl":
            encode_observation(self.memory_systems, self.policies["encoding"], self.obs)
            manage_memory(self.memory_systems, self.policies["memory_management"])
            while True:
                if (self.question is None) and (self.answer is None):
                    (
                        self.obs,
                        self.question,
                        self.answer,
                        self.is_last,
                    ) = self.generate_oqa(increment_des=True)
                    encode_observation(
                        self.memory_systems, self.policies["encoding"], self.obs
                    )
                    manage_memory(
                        self.memory_systems, self.policies["memory_management"]
                    )
                else:
                    return {
                        "memory_systems": deepcopy(
                            self.extract_memory_entires(self.memory_systems)
                        ),
                        "question": deepcopy(self.question),
                    }, info

        raise ValueError

    def step(self, action: int) -> Tuple[Tuple, int, bool, bool, dict]:
        """An agent takes an action.

        Args
        ----
        action: This depends on the state

        Returns
        -------
        state, reward, done, truncated, info

        """
        info = {}
        truncated = False
        if self.policies["encoding"].lower() == "rl":
            # This is a dummy code
            self.obs = self.obs[action]
            encode_observation(self.memory_systems, self.policies["encoding"], self.obs)
            manage_memory(self.memory_systems, self.policies["memory_management"])

            if (self.question is None) and (self.answer is None):
                reward = 0
            else:
                pred = answer_question(
                    self.memory_systems, self.policies["question_answer"], self.question
                )
                if str(pred).lower() == self.answer:
                    reward = self.CORRECT
                else:
                    reward = self.WRONG
            self.obs, self.question, self.answer, self.is_last = self.generate_oqa(
                increment_des=True
            )
            state = deepcopy(self.obs)

            if self.is_last:
                done = True
            else:
                done = False

            return state, reward, done, truncated, info

        if self.policies["memory_management"].lower() == "rl":
            if action == 0:
                manage_memory(self.memory_systems, "episodic")
            elif action == 1:
                manage_memory(self.memory_systems, "semantic")
            elif action == 2:
                manage_memory(self.memory_systems, "forget")
            else:
                raise ValueError

            if (self.question is None) and (self.answer is None):
                reward = 0
            else:
                pred = answer_question(
                    self.memory_systems, self.policies["question_answer"], self.question
                )
                if str(pred).lower() == self.answer:
                    reward = self.CORRECT
                else:
                    reward = self.WRONG

            self.obs, self.question, self.answer, self.is_last = self.generate_oqa(
                increment_des=True
            )
            encode_observation(self.memory_systems, self.policies["encoding"], self.obs)
            state = deepcopy(self.extract_memory_entires(self.memory_systems))

            if self.is_last:
                done = True
            else:
                done = False

            return state, reward, done, truncated, info

        if self.policies["question_answer"].lower() == "rl":
            if action == 0:
                pred = answer_question(self.memory_systems, "episodic", self.question)
            elif action == 1:
                pred = answer_question(self.memory_systems, "semantic", self.question)
            else:
                raise ValueError

            if str(pred).lower() == self.answer:
                reward = self.CORRECT
            else:
                reward = self.WRONG

            while True:
                (
                    self.obs,
                    self.question,
                    self.answer,
                    self.is_last,
                ) = self.generate_oqa(increment_des=True)
                encode_observation(
                    self.memory_systems, self.policies["encoding"], self.obs
                )
                manage_memory(self.memory_systems, self.policies["memory_management"])

                if self.is_last:
                    state = None
                    done = True
                    return state, reward, done, truncated, info
                else:
                    done = False

                if (self.question is not None) and (self.answer is not None):
                    state = {
                        "memory_systems": deepcopy(
                            self.extract_memory_entires(self.memory_systems)
                        ),
                        "question": deepcopy(self.question),
                    }

                    return state, reward, done, truncated, info

    def render(self, mode="console") -> None:
        if mode != "console":
            raise NotImplementedError()
        else:
            pass



import unittest

from room_env.memory import EpisodicMemory, Memory, SemanticMemory, ShortMemory
from room_env.policy import answer_question, encode_observation, manage_memory


class PolicyTest(unittest.TestCase):
    def setUp(self) -> None:
        self.memory_systems = {
            "short": ShortMemory(capacity=1),
            "episodic": EpisodicMemory(capacity=16),
            "semantic": SemanticMemory(capacity=16),
        }

    def test_encode_observation(self):
        with self.assertRaises(NotImplementedError):
            encode_observation(
                memory_systems=self.memory_systems,
                policy="foo",
                obs={"human": "Tae", "object": "Ubuntu", "object_location": "Linux"},
            )

    def test_manage_memory(self):
        with self.assertRaises(NotImplementedError):
            manage_memory(memory_systems=self.memory_systems, policy="neural")

    def test_answer_question(self):
        with self.assertRaises(NotImplementedError):
            answer_question(
                memory_systems=self.memory_systems,
                policy="neural",
                question={"human": "Tae", "object": "Ubuntu"},
            )



import unittest

from room_env.memory import EpisodicMemory, Memory, SemanticMemory, ShortMemory


class MemoryTest(unittest.TestCase):
    def test_all(self) -> None:
        for capacity in [1, 2, 4, 8, 16, 32]:
            m = EpisodicMemory(capacity=capacity)
            m = SemanticMemory(capacity=capacity)
            m = ShortMemory(capacity=capacity)



"""Utility functions"""
import json
import logging
import os
import random
import subprocess
from copy import deepcopy
from typing import List, Tuple

import gym
import numpy as np
import torch
import yaml

import room_env

from .des import RoomDes

logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def seed_everything(seed: int) -> None:
    """Seed every randomness to seed"""
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


def read_lines(fname: str) -> list:
    """Read lines from a text file.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    if fname.startswith("/"):
        fullpath = fname
    else:
        fullpath = os.path.join(os.path.dirname(__file__), fname)

    logging.debug(f"Reading {fullpath} ...")
    with open(fullpath, "r") as stream:
        names = stream.readlines()
    names = [line.strip() for line in names]

    return names


def read_json(fname: str) -> dict:
    """Read json"""
    logging.debug(f"reading json {fname} ...")
    with open(fname, "r") as stream:
        return json.load(stream)


def write_json(content: dict, fname: str) -> None:
    """Write json"""
    logging.debug(f"writing json {fname} ...")
    with open(fname, "w") as stream:
        json.dump(content, stream, indent=4, sort_keys=False)


def read_yaml(fname: str) -> dict:
    """Read yaml.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    if fname.startswith("/"):
        fullpath = fname
    else:
        fullpath = os.path.join(os.path.dirname(__file__), fname)
    logging.debug(f"reading yaml {fullpath} ...")
    with open(fullpath, "r") as stream:
        return yaml.safe_load(stream)


def write_yaml(content: dict, fname: str) -> None:
    """write yaml."""
    logging.debug(f"writing yaml {fname} ...")
    with open(fname, "w") as stream:
        yaml.dump(content, stream, indent=2, sort_keys=False)


def read_data(data_path: str) -> dict:
    """Read train, val, test spilts.

    Args
    ----
    data_path: path to data.

    Returns
    -------
    data: {'train': list of training obs,
           'val': list of val obs,
           'test': list of test obs}
    """
    logging.debug(f"reading data from {data_path} ...")
    data = read_json(data_path)
    logging.info(f"Succesfully read data {data_path}")

    return data


def argmax(iterable):
    """argmax"""
    return max(enumerate(iterable), key=lambda x: x[1])[0]


def remove_name(entity: str) -> str:
    """Remove name from the entity.

    Args
    ----
    entity: e.g., Bob's laptop

    Returns
    -------
    e.g., laptop

    """
    return entity.split()[-1]


def split_name_entity(name_entity: str) -> Tuple[str, str]:
    """Separate name and entity from the given string.

    Args
    ----
    name_entity: e.g., "Bob's laptop"

    Returns
    -------
    name: e.g., Bob
    entity: e.g., laptop

    """
    logging.debug(f"spliting name and entity from {name_entity}")
    splitted = name_entity.split()
    assert len(splitted) == 2 and "'" in splitted[0]
    name = splitted[0].split("'")[0]
    entity = splitted[1]

    return name, entity


def get_duplicate_dicts(search: dict, target: list) -> List:
    """Find if there are duplicate dicts.

    Args
    ----
    search: dict
    target: target list to look up.

    Returns
    -------
    duplicates: a list of dicts or None

    """
    assert isinstance(search, dict)
    logging.debug("finding if duplicate dicts exist ...")
    duplicates = []

    for candidate in target:
        assert isinstance(candidate, dict)
        if set(search).issubset(set(candidate)):
            if all([val == candidate[key] for key, val in search.items()]):
                duplicates.append(candidate)

    logging.info(f"{len(duplicates)} duplicates were found!")

    return duplicates


def list_duplicates_of(seq, item) -> List:
    # https://stackoverflow.com/questions/5419204/index-of-duplicates-items-in-a-python-list
    start_at = -1
    locs = []
    while True:
        try:
            loc = seq.index(item, start_at + 1)
        except ValueError:
            break
        else:
            locs.append(loc)
            start_at = loc
    return locs


def make_des_config(
    commonsense_prob: float,
    num_humans: int,
    num_total_objects: int,
    maximum_num_objects_per_human: int,
    maximum_num_locations_per_object: int,
    maxiumum_days_period: int,
    des_size: str,
    last_timestep: int = 128,
) -> dict:
    """Make a des config.

    Args
    ----
    commonsense_prob: commonsense probability
    num_humans: number of humans
    num_total_objects: number of total objects
    maximum_num_objects_per_human: maximum number of objects per human
    maximum_num_locations_per_object: maximum number of locations per object
    maxiumum_days_period: maxiumum number of days period
    des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")
    last_timestep: last time step where the DES terminates.

    Returns
    -------
    des config

    """
    des_config = {
        "human_names_path": "./room_env/data/human-names",
        "last_timestep": last_timestep,
        "maxiumum_days_period": maxiumum_days_period,
        "save_path": f"./room_env/data/des-config-{des_size}.json",
        "seed": 42,
        "semantic_knowledge_path": "./room_env/data/semantic-knowledge.json",
    }

    des_config["commonsense_prob"] = commonsense_prob
    des_config["num_humans"] = num_humans
    des_config["num_total_objects"] = num_total_objects
    des_config["maximum_num_objects_per_human"] = maximum_num_objects_per_human
    des_config["maximum_num_locations_per_object"] = maximum_num_locations_per_object

    return des_config


def get_des_variables(des_size: str = "l") -> Tuple[int, int, int]:
    """Get the des variables.

    Args
    ----
    des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")

    Returns
    -------
    capacity, num_humans, num_total_objects

    """
    if des_size == "dev":
        capacity = 16

    elif des_size == "xxs":
        capacity = 2

    elif des_size == "xs":
        capacity = 4

    elif des_size == "s":
        capacity = 8

    elif des_size == "m":
        capacity = 16

    elif des_size == "l":
        capacity = 32

    else:
        raise ValueError

    num_humans = capacity * 2
    num_total_objects = capacity // 2

    return capacity, num_humans, num_total_objects


def run_des_seeds(
    seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    capacity: int = 32,
    des_size: str = "l",
    allow_random_human: bool = True,
    allow_random_question: bool = True,
    question_prob: float = 0.1,
) -> dict:
    """Run the RoomEnv-v1 with multiple different seeds.

    Args
    ----
    seeds:
    capacity:
    des_size:
    allow_random_human:
    allow_random_question:
    question_prob:

    Returns
    -------
    results

    """
    results = {}
    how_to_forget = ["episodic", "semantic", "random", "pre_sem"]

    for forget_short in how_to_forget:

        if forget_short == "random":
            pretrain_semantic = False
            capacity_ = {
                "episodic": capacity // 2,
                "semantic": capacity // 2,
                "short": 1,
            }
        elif forget_short == "episodic":
            pretrain_semantic = False
            capacity_ = {"episodic": capacity, "semantic": 0, "short": 1}
        elif forget_short == "semantic":
            pretrain_semantic = False
            capacity_ = {"episodic": 0, "semantic": capacity, "short": 1}
        elif forget_short == "pre_sem":
            pretrain_semantic = True
            capacity_ = {
                "episodic": capacity // 2,
                "semantic": capacity // 2,
                "short": 1,
            }
        else:
            raise ValueError

        results_ = []
        for seed in seeds:
            env = gym.make(
                "RoomEnv-v1",
                des_size=des_size,
                seed=seed,
                policies={
                    "memory_management": "rl",
                    "question_answer": "episodic_semantic",
                    "encoding": "argmax",
                },
                capacity=capacity_,
                question_prob=question_prob,
                observation_params="perfect",
                allow_random_human=allow_random_human,
                allow_random_question=allow_random_question,
                pretrain_semantic=pretrain_semantic,
                check_resources=False,
                varying_rewards=False,
            )
            state, info = env.reset()
            rewards = 0
            while True:
                if forget_short == "random":
                    action = random.choice([0, 1, 2])
                elif forget_short == "episodic":
                    action = 0
                elif forget_short == "semantic":
                    action = 1
                elif forget_short == "pre_sem":
                    action = 0
                else:
                    raise ValueError
                state, reward, done, truncated, info = env.step(action)
                rewards += reward
                if done:
                    break
            results_.append(rewards)

        results[forget_short] = np.mean(results_)

    return results


def run_all_des_configs(
    des_size: str,
    capacity: int,
    maximum_num_objects_per_human: int,
    maximum_num_locations_per_object: int,
    maxiumum_days_period: int,
    commonsense_prob: float,
    num_humans: int,
    num_total_objects: int,
    seeds: list,
    allow_random_human: bool,
    allow_random_question: bool,
    last_timestep: int,
    question_prob: float,
) -> dict:
    """Run the RoomEnv-v1 with different des configs, with multiple different seeds.

    Args
    ----
    des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")
    capacity: int,
    maximum_num_objects_per_human: maximum number of objects per human
    maximum_num_locations_per_object: maximum number of locations per object
    maxiumum_days_period: maxiumum number of days period
    commonsense_prob: commonsense probability
    num_humans: number of humans
    num_total_objects: number of total objects
    seeds: list,
    allow_random_human: bool,
    allow_random_question: bool,
    last_timestep: int,
    question_prob: float,

    Returns
    -------
    results

    """
    des_config = make_des_config(
        commonsense_prob=commonsense_prob,
        num_humans=num_humans,
        num_total_objects=num_total_objects,
        maximum_num_objects_per_human=maximum_num_objects_per_human,
        maximum_num_locations_per_object=maximum_num_locations_per_object,
        maxiumum_days_period=maxiumum_days_period,
        des_size=des_size,
        last_timestep=last_timestep,
    )

    complexity = (
        num_humans
        * num_total_objects
        * maximum_num_objects_per_human
        * maximum_num_locations_per_object
        * maxiumum_days_period
    )

    with open("create_des_config.yaml", "w") as stream:
        yaml.safe_dump(des_config, stream, indent=2)

    sub_out = subprocess.call(
        ["python", "create_des_config.py"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if sub_out == 1:
        return None

    rewards = run_des_seeds(
        seeds=seeds,
        capacity=capacity,
        des_size=des_size,
        allow_random_human=allow_random_human,
        allow_random_question=allow_random_question,
        question_prob=question_prob,
    )
    results = {
        "mean_rewards_diff": rewards["pre_sem"]
        - rewards["random"] / 3
        - rewards["semantic"] / 3
        - rewards["episodic"] / 3,
        "mean_rewards_episodic": rewards["episodic"],
        "mean_rewards_semantic": rewards["semantic"],
        "mean_rewards_random": rewards["random"],
        "mean_rewards_pre_sem": rewards["pre_sem"],
        "complexity": complexity,
        "commonsense_prob": commonsense_prob,
        "maximum_num_locations_per_object": maximum_num_locations_per_object,
        "maximum_num_objects_per_human": maximum_num_objects_per_human,
        "num_humans": num_humans,
        "num_total_objects": num_total_objects,
        "maxiumum_days_period": maxiumum_days_period,
        "allow_random_human": allow_random_human,
        "allow_random_question": allow_random_question,
        "question_prob": question_prob,
    }
    return deepcopy(results)


def fill_des_resources(des_size: str) -> None:
    """Fill resources

    Args
    ----
    des_size

    """
    des = RoomDes(des_size=des_size, check_resources=False)
    des.run()
    resources = {
        foo: 9999
        for foo in set(
            [bar["object_location"] for foo in des.states for bar in foo.values()]
        )
    }
    des.config["resources"] = deepcopy(resources)
    write_json(des.config, f"./room_env/data/des-config-{des_size}.json")
    resources = []
    des = RoomDes(des_size=des_size, check_resources=True)
    resources.append(deepcopy(des.resources))
    while des.until > 0:
        des.step()
        des.until -= 1
        resources.append(deepcopy(des.resources))

    object_locations = deepcopy(list(des.resources.keys()))
    resources = {
        object_location: 9999
        - min([resource[object_location] for resource in resources])
        for object_location in object_locations
    }

    des.config["resources"] = deepcopy(resources)
    write_json(des.config, f"./room_env/data/des-config-{des_size}.json")
    des = RoomDes(des_size=des_size, check_resources=True)


def get_handcrafted(
    env: str = "RoomEnv-v1",
    des_size: str = "l",
    seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    question_prob: float = 0.1,
    observation_params: str = "perfect",
    policies: dict = {
        "memory_management": "rl",
        "question_answer": "episodic_semantic",
        "encoding": "argmax",
    },
    capacities: list = [2, 4, 8, 16, 32, 64],
    allow_random_human: bool = False,
    allow_random_question: bool = True,
    varying_rewards: bool = False,
    check_resources: bool = True,
) -> None:
    """Get the env results with handcrafted policies.

    At the moment only {"memory_management": "rl"} is supported.

    Args
    ----
    env: str = "RoomEnv-v1",
    des_size: str = "l",
    seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    question_prob: float = 0.1,
    policies: dict = {
        "memory_management": "rl",
        "question_answer": "episodic_semantic",
        "encoding": "argmax",
    },
    capacities: list = [2, 4, 8, 16, 32, 64],
    allow_random_human: whether to allow random humans to be observed.
    allow_random_question: whether to allow random questions to be asked.
    varying_rewards: If true, then the rewards are scaled in every episode so that
            total_episode_rewards is 128.

    Returns
    -------
    handcrafted_results

    """
    how_to_forget = ["episodic", "semantic", "random", "pre_sem"]
    env_ = env
    handcrafted_results = {}

    for capacity in capacities:
        handcrafted_results[capacity] = {}
        for forget_short in how_to_forget:

            if forget_short == "random":
                pretrain_semantic = False
                capacity_ = {
                    "episodic": capacity // 2,
                    "semantic": capacity // 2,
                    "short": 1,
                }
            elif forget_short == "episodic":
                pretrain_semantic = False
                capacity_ = {"episodic": capacity, "semantic": 0, "short": 1}
            elif forget_short == "semantic":
                pretrain_semantic = False
                capacity_ = {"episodic": 0, "semantic": capacity, "short": 1}
            elif forget_short == "pre_sem":
                pretrain_semantic = True
                capacity_ = {
                    "episodic": capacity // 2,
                    "semantic": capacity // 2,
                    "short": 1,
                }
            else:
                raise ValueError

            results = []
            for seed in seeds:
                env = gym.make(
                    env_,
                    des_size=des_size,
                    seed=seed,
                    policies=policies,
                    capacity=capacity_,
                    question_prob=question_prob,
                    observation_params=observation_params,
                    allow_random_human=allow_random_human,
                    allow_random_question=allow_random_question,
                    pretrain_semantic=pretrain_semantic,
                    check_resources=check_resources,
                    varying_rewards=varying_rewards,
                )
                state, info = env.reset()
                rewards = 0
                while True:
                    if forget_short == "random":
                        action = random.choice([0, 1, 2])
                    elif forget_short == "episodic":
                        action = 0
                    elif forget_short == "semantic":
                        action = 1
                    elif forget_short == "pre_sem":
                        action = 0
                    else:
                        raise ValueError
                    state, reward, done, truncated, info = env.step(action)
                    rewards += reward
                    if done:
                        break
                results.append(rewards)

            mean_ = np.mean(results).round(3).item()
            std_ = np.std(results).round(3).item()
            handcrafted_results[capacity][forget_short] = {"mean": mean_, "std": std_}

    return handcrafted_results



"""Room environment compatible with gym.
I advise you to check out room1.py, as it's more general.
"""
import json
import logging
import os
import random
from copy import deepcopy
from itertools import cycle
from typing import Tuple

import gym

from ..utils import read_lines, remove_name, split_name_entity

CORRECT = 1
WRONG = 0

logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def read_json(fname: str) -> dict:
    """Read json.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    fullpath = os.path.join(os.path.dirname(__file__), "../", fname)

    with open(fullpath, "r") as stream:
        return json.load(stream)


class RoomEnv0(gym.Env):
    """The Room environment version 0.

    In this big room, N_{agents} move around and observe N_{humans} placing objects.
    Every time the agents move around (i.e., take a step), they observe one human_{i}
    placing an object somewhere. Each agent can only observe one human at a time.
    Every time the agents takes a step, the environment also asks them where an object
    is. +1 reward is given when it gets right and 0 when it gets wrong.

    This environment is challenging in two ways:

    (1) An agent can't observe the entire room. It can only observe one human at a time.
    This means that the environment is only partially observable. This constraint means
    that it's more beneficial if more than one agent can collaborate. This constraint
    also means that the agents should have a memory system to remember the past
    observations.

    (2) The room environment changes every time. The humans can switch their locations,
    change their objects, and place them at different locations. These changes are
    not completely random. A decent portion of them come from commmon-sense knowledge.
    This means that an agent with both episodic and semantic memory systems will perform
    better than an agent with only one memory system.

    """

    metadata = {"render.modes": ["console"]}

    def __init__(
        self,
        room_size: str = "small",
        weighting_mode: str = "highest",
        probs: dict = {
            "commonsense": 0.7,
            "new_location": 0.1,
            "new_object": 0.1,
            "switch_person": 0.5,
        },
        limits: dict = {
            "heads": None,
            "tails": None,
            "names": None,
            "allow_spaces": False,
        },
        max_step: int = 1000,
        disjoint_entities: bool = True,
        num_agents: int = 1,
    ) -> None:
        """Initialize the environment.

        Args
        ----
        room_size: small or big
        weighting_mode: Either "weighted" or "highest"
        probs: the probabilities that govern the room environment changes.
        limits: Limitation on the triples.
        max_step: maximum step an agent can take. The environment will terminate when
            the number reaches this value.
        disjoint_entities: Assure that the heads and the tails don't overlap.
        num_agents: number of agents in the room.

        """
        super().__init__()
        self.room_size = room_size
        if self.room_size.lower() == "small":
            semantic_knowledge_path = "./data/semantic-knowledge-small.json"
            names_path = "./data/top-human-names-small"

        elif self.room_size.lower() in ["large", "big"]:
            semantic_knowledge_path = "./data/semantic-knowledge.json"
            names_path = "./data/top-human-names"

        logging.debug("Creating an Observation-Question-Answer generator object ...")
        self.limits = limits

        if set(list(self.limits.values())) != {None}:
            logging.warning(f"The obserations will be limited by {self.limits}")

        (
            self.semantic_knowledge,
            self.heads,
            self.relations,
            self.tails,
        ) = self.load_semantic_knowledge(
            semantic_knowledge_path,
            limit_heads=self.limits["heads"],
            limit_tails=self.limits["tails"],
            allow_spaces=self.limits["allow_spaces"],
            disjoint_entities=disjoint_entities,
        )

        assert len(self.relations) == 1, "At the moment there is only one relation."

        self.names = self.read_names(
            names_path, self.limits["names"], self.limits["allow_spaces"]
        )

        assert len(self.names) <= len(self.heads)

        if disjoint_entities:
            lhs = len(set(self.relations + self.names + self.heads + self.tails))
            rhs = (
                len(self.relations)
                + len(self.names)
                + len(self.heads)
                + len(self.tails)
            )

            assert lhs == rhs

        self.weighting_mode = weighting_mode
        self.probs = probs
        self.max_step = max_step
        self.num_agents = num_agents
        # Our state / action space is quite complex. Here we just make a dummy
        # observation space to bypass the sanity check.
        self.observation_space = gym.spaces.Discrete(1)
        self.action_space = gym.spaces.Discrete(1)

    def reset(self) -> Tuple:
        """Reset the environment.

        This will place N_{humans} humans in the room. Each human only has one object,
        which will be placed by the human in a random location.

        """
        self.step_counter = 0
        random.shuffle(self.names)
        random.shuffle(self.heads)
        self.room = []

        for name, head in zip(self.names, self.heads):
            relation = self.relations[0]  # At the moment there is only one relation.
            tail = self.generate_tail(head, relation)
            if tail is not None:
                self.room.append([f"{name}'s {head}", relation, f"{name}'s {tail}"])

        navigate = [[i for i in range(len(self.room))] for _ in range(self.num_agents)]
        for navigate_ in navigate:
            random.shuffle(navigate_)

        self.navigate = [cycle(navigate_) for navigate_ in navigate]

        observations = self.generate_observations()
        question, self.answer = self.generate_qa()

        info = {}

        return (observations, question), info

    def generate_observations(self) -> list:
        """Generate a random obseration.

        Returns
        -------
        observations: e.g., ["Tae's laptop, "AtLocation", "Tae's desk", 10]

        The last element in the list accounts for the timestamp.

        """
        observations = {
            i: deepcopy([*self.room[next(navigate_)], self.step_counter])
            for i, navigate_ in enumerate(self.navigate)
        }

        return observations

    def generate_qa(self) -> Tuple[list, str]:
        """Generate a question and the answer.

        Returns
        -------
        question: e.g., ["Tae's laptop", "AtLocation"]
        answer: e.g., "Tae's desk"

        """
        random_choice = random.choice(self.room)
        question = random_choice[:2]
        answer = remove_name(random_choice[-1])

        return question, answer

    def generate_tail(self, head: str, relation: str) -> str:
        """This simulates humans placing their objects in their desired locations.

        Note that "head" shouldn't include a human name.

        """
        if random.random() < self.probs["commonsense"]:
            logging.debug(f"Generating a common location for {head} ...")
            tails = self.semantic_knowledge[head][relation]

            if len(tails) == 0:
                return None

            if self.weighting_mode == "weighted":
                tail = random.choices(
                    [tail["tail"] for tail in tails],
                    weights=[tail["weight"] for tail in tails],
                    k=1,
                )[0]
            elif self.weighting_mode == "highest":
                tail = sorted(
                    self.semantic_knowledge[head][relation], key=lambda x: x["weight"]
                )[-1]["tail"]
            else:
                raise ValueError
        else:
            logging.debug(f"Generating a NON common location for {head} ...")
            while True:
                tail = random.choice(self.tails)
                if tail not in self.semantic_knowledge[head][relation]:
                    break

        return tail

    def renew(self) -> None:
        """Renew the room.

        This is done every time when the agent takes a step. This is doen to simulate
        that the room changes. People move. They place their objects in different
        locations. They also change their objects. All of these are done in a random
        manner. The randomness can be adjusted from the argument `probs`.

        With the chance of probs["new_location"], an object will be placed at a new
        location.

        When the object is placed at a new locaiton, with the chance of
        probs["commonsense"], an object will be placed at a commonsense-knowledge spot.

        With the chance of probs["new_object"], the person with the object will have
        a new random object.

        With the chance of probs["switch_person"], two persons switch their spots.

        """
        room = []
        for head, relation, tail in self.room:
            name1, head = split_name_entity(head)
            name2, tail = split_name_entity(tail)

            assert name1 == name2, "we don't do name mixing at this moment."

            if random.random() < self.probs["new_object"]:
                while True:
                    new_head = random.choice(self.heads)
                    if new_head != head:
                        head = new_head
                        tail = self.generate_tail(head, relation)
                        break
            else:
                if random.random() < self.probs["new_location"]:
                    while True:
                        new_tail = self.generate_tail(head, relation)
                        if new_tail != tail:
                            tail = new_tail
                            break

            room.append(
                [
                    f"{name1}'s {deepcopy(head)}",
                    deepcopy(relation),
                    f"{name2}'s {deepcopy(tail)}",
                ],
            )

        if random.random() < self.probs["switch_person"]:
            i, j = random.sample(
                range(
                    0,
                    len(
                        self.room,
                    ),
                ),
                2,
            )
            room[i], room[j] = room[j], room[i]

        self.room = room

    def step(self, action: str) -> Tuple[Tuple, int, bool, bool, dict]:
        """An agent takes an action.

        Args
        ----
        action: This is the agent's answer to the previous question.

        """
        if str(action).lower() == self.answer.lower():
            logging.info(
                f"The prediction ({action}) matches the answer ({self.answer})!"
            )
            reward = CORRECT

        else:
            logging.info(
                f"The prediction ({action}) does NOT match the answer ({self.answer})!"
            )
            reward = WRONG

        self.step_counter += 1

        # Things will change in the room!
        self.renew()

        observations = self.generate_observations()
        question, self.answer = self.generate_qa()

        info = {}

        if self.step_counter >= self.max_step:
            done = True
        else:
            done = False

        truncated = False

        return (observations, question), reward, done, truncated, info

    def render(self, mode="console") -> None:
        if mode != "console":
            raise NotImplementedError()
        else:
            pass

    def close(self):
        pass

    @staticmethod
    def read_names(
        path: str = "./data/top-human-names",
        limit_names: int = None,
        allow_spaces: bool = False,
    ) -> list:
        """Read 20 most common names.

        Args
        ----
        path: The path to the top 20 human name list.
        limit_names: Limit the number of names
        allow_spaces: Whether to include words that have spaces
            (e.g., corner of two streets)

        Returns
        -------
        names: human names (e.g., James)

        """
        names = read_lines(path)

        if not allow_spaces:
            names = [name for name in names if len(name.split("_")) == 1]

        if limit_names:
            logging.warning(f"The number of names will be limited to {limit_names}")
            names = sorted(names, key=len)
            names = names[:limit_names]

        logging.info(f"Reading {path} complete! There are {len(names)} names in total")

        return names

    @staticmethod
    def load_semantic_knowledge(
        path: str,
        limit_heads: int = None,
        limit_tails: int = None,
        allow_spaces: bool = False,
        disjoint_entities: bool = True,
    ) -> Tuple[list, list, list, list]:
        """Load saved semantic knowledge.

        Args
        ----
        path: the path to the pretrained semantic memory.
        limit_heads: Limit the number of heads (e.g., 10)
        limit_tails: Limit the number of tails per heads (e.g., 1)
        allow_spaces: Whether to include words that have spaces
            (e.g., corner of two streets)
        disjoint_entities: Whether to force that there are no common elements between
            entities.

        Returns
        -------
        semantic_knowledge
        heads
        relations
        tails

        """
        logging.debug(f"loading the semantic knowledge from {path}...")
        semantic_knowledge = read_json(path)

        heads = sorted(list(set(semantic_knowledge.keys())))
        if disjoint_entities:
            logging.warning("Tails that are heads will be removed.")
            semantic_knowledge = {
                key: {
                    key_: [tail for tail in val_ if tail["tail"] not in heads]
                    for key_, val_ in val.items()
                }
                for key, val in semantic_knowledge.items()
            }

        semantic_knowledge = {
            key: {
                key_: val_
                for key_, val_ in val.items()
                if len([tail for tail in val_ if tail["tail"]]) > 0
            }
            for key, val in semantic_knowledge.items()
        }
        semantic_knowledge = {
            key: val for key, val in semantic_knowledge.items() if len(val) > 0
        }
        logging.info("empty entities are removed")

        # sort the semantic knowledge by its highest weight to be sure.
        semantic_knowledge = {
            key: {
                key_: sorted(val_, key=lambda x: -x["weight"])
                for key_, val_ in val.items()
            }
            for key, val in semantic_knowledge.items()
        }

        if not allow_spaces:
            semantic_knowledge = {
                key: {
                    key_: [v for v in val_ if len(v["tail"].split("_")) == 1]
                    for key_, val_ in val.items()
                }
                for key, val in semantic_knowledge.items()
                if (len(key.split("_"))) == 1
            }

        if limit_heads:
            logging.warning(f"Limiting the number of heads to {limit_heads} ...")
            semantic_knowledge = {
                key: val
                for idx, (key, val) in enumerate(semantic_knowledge.items())
                if idx < limit_heads
            }

        if limit_tails:
            logging.warning(
                f"Limiting the number of tails per head to {limit_tails} ..."
            )
            semantic_knowledge = {
                key: {key_: val_[:limit_tails] for key_, val_ in val.items()}
                for key, val in semantic_knowledge.items()
            }

        heads = sorted(list(set(semantic_knowledge.keys())))
        tails = sorted(
            list(
                set(
                    [
                        foo["tail"]
                        for key, val in semantic_knowledge.items()
                        for key_, val_ in val.items()
                        for foo in val_
                    ]
                )
            )
        )
        relations = sorted(
            list(
                set(
                    [
                        key_
                        for key, val in semantic_knowledge.items()
                        for key_, val_ in val.items()
                    ]
                )
            )
        )
        logging.info(f"semantic knowledge successfully loaded from {path}!")

        return semantic_knowledge, heads, relations, tails



import unittest

import gym

import room_env


class RoomEnv1Test(unittest.TestCase):
    def test_all(self) -> None:
        for room_size in ["small", "big"]:
            env = gym.make("RoomEnv-v0", room_size=room_size)
            observations, info = env.reset()
            while True:
                observations, reward, done, truncated, info = env.step("foo")
                if done:
                    break



import logging
import random
import unittest

import gym

import room_env

logger = logging.getLogger()
logger.disabled = True


class RoomEnv1Test(unittest.TestCase):
    def test_all(self) -> None:
        for des_size in ["xxs", "xs", "s", "m", "l"]:
            for question_prob in [0.1, 0.2, 0.4, 1]:
                for allow_random_human in [True, False]:
                    for allow_random_question in [True, False]:
                        for pretrain_semantic in [True, False]:
                            for check_resources in [True, False]:
                                for varying_rewards in [True, False]:
                                    env = gym.make(
                                        "RoomEnv-v1",
                                        des_size=des_size,
                                        question_prob=question_prob,
                                        allow_random_human=allow_random_human,
                                        allow_random_question=allow_random_question,
                                        pretrain_semantic=pretrain_semantic,
                                        check_resources=check_resources,
                                        varying_rewards=varying_rewards,
                                    )
                                    state, info = env.reset()
                                    while True:
                                        (
                                            state,
                                            reward,
                                            done,
                                            truncated,
                                            info,
                                        ) = env.step(random.randint(0, 2))
                                        if done:
                                            break

    def test_wrong_init0(self) -> None:

        with self.assertRaises(AssertionError):
            env = gym.make(
                "RoomEnv-v1",
                policies={
                    "memory_management": "RL",
                    "question_answer": "RL",
                    "encoding": "argmax",
                },
            )
            del env

    def test_wrong_init1(self) -> None:

        with self.assertRaises(NotImplementedError):
            env = gym.make(
                "RoomEnv-v1",
                policies={
                    "memory_management": "episodic",
                    "question_answer": "episodic",
                    "encoding": "rl",
                },
            )
            del env

    def test_wrong_init2(self) -> None:

        with self.assertRaises(NotImplementedError):
            env = gym.make("RoomEnv-v1", observation_params="foo")
            state, info = env.reset()
            del env

    def test_rewards(self) -> None:
        env = gym.make(
            "RoomEnv-v1",
            policies={
                "memory_management": "RL",
                "question_answer": "episodic_semantic",
                "encoding": "argmax",
            },
            seed=random.randint(0, 10000),
            question_prob=0.1,
            capacity={"episodic": 16, "semantic": 16, "short": 1},
            varying_rewards=True,
        )
        state, info = env.reset()
        self.assertAlmostEqual(
            env.total_episode_rewards, env.CORRECT * env.num_questions
        )

    def test_reset_qa(self) -> None:
        for memory_management in ["episodic", "semantic"]:
            env = gym.make(
                "RoomEnv-v1",
                policies={
                    "memory_management": memory_management,
                    "question_answer": "rl",
                    "encoding": "argmax",
                },
                seed=random.randint(0, 10000),
                question_prob=0.1,
                capacity={"episodic": 16, "semantic": 16, "short": 1},
            )
            state, info = env.reset()
            self.assertIn("memory_systems", state)
            self.assertIn("question", state)

            self.assertIn("episodic", state["memory_systems"])
            self.assertIn("semantic", state["memory_systems"])
            self.assertIn("short", state["memory_systems"])
            self.assertIn("question", state)

            if memory_management == "semantic":
                self.assertGreater(len(state["memory_systems"]["semantic"]), 0)
                self.assertEqual(len(state["memory_systems"]["episodic"]), 0)

            else:
                self.assertEqual(len(state["memory_systems"]["semantic"]), 0)
                self.assertGreater(len(state["memory_systems"]["episodic"]), 0)

            self.assertEqual(len(state["memory_systems"]["short"]), 0)

            while True:
                state, reward, done, truncated, info = env.step(random.randint(0, 1))
                if done:
                    break
                self.assertIn("episodic", state["memory_systems"])
                self.assertIn("semantic", state["memory_systems"])
                self.assertIn("short", state["memory_systems"])
                self.assertIn("question", state)

    def test_reset_memory_management(self) -> None:
        for qa_policy in ["episodic", "semantic", "episodic_semantic"]:
            env = gym.make(
                "RoomEnv-v1",
                policies={
                    "memory_management": "rl",
                    "question_answer": qa_policy,
                    "encoding": "argmax",
                },
                seed=random.randint(0, 10000),
                question_prob=0.1,
                capacity={"episodic": 16, "semantic": 16, "short": 1},
            )
            state, info = env.reset()
            self.assertIn("episodic", state)
            self.assertIn("semantic", state)
            self.assertIn("short", state)

            self.assertEqual(len(state["semantic"]), 0)
            self.assertEqual(len(state["episodic"]), 0)
            self.assertEqual(len(state["short"]), 1)
            while True:
                state, reward, done, truncated, info = env.step(random.randint(0, 2))
                if done:
                    break
                self.assertIn("episodic", state)
                self.assertIn("semantic", state)
                self.assertIn("short", state)



"""Memory system classes."""
import logging
import os
import random
from copy import deepcopy
from pprint import pformat
from typing import List, Tuple

from .utils import get_duplicate_dicts, list_duplicates_of

logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Memory:
    """Memory (episodic, semantic, or short) class"""

    def __init__(self, memory_type: str, capacity: int) -> None:
        """

        Args
        ----
        memory_type: episodic, semantic, or short
        capacity: memory capacity

        """
        logging.debug(
            f"instantiating a {memory_type} memory object with size {capacity} ..."
        )

        assert memory_type in ["episodic", "semantic", "short"]
        self.type = memory_type
        self.entries = []
        self.capacity = capacity
        self._frozen = False

        logging.debug(f"{memory_type} memory object with size {capacity} instantiated!")

    def __repr__(self):

        return pformat(vars(self), indent=4, width=1)

    def forget(self, mem: dict) -> None:
        """forget the given memory.

        Args
        ----
        mem: A memory in a dictionary format.

            for episodic and short:
            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

            for semantic:
            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        if self._frozen:
            error_msg = "The memory system is frozen!"
            logging.error(error_msg)
            raise ValueError(error_msg)

        if mem not in self.entries:
            error_msg = f"{mem} is not in the memory system!"
            logging.error(error_msg)
            raise ValueError(error_msg)

        logging.debug(f"Forgetting {mem} ...")
        self.entries.remove(mem)
        logging.info(f"{mem} forgotten!")

    def forget_all(self) -> None:
        """Forget everything in the memory system!"""
        if self.is_frozen:
            logging.warning(
                "The memory system is frozen. Can't forget all. Unfreeze first."
            )
        else:
            logging.warning("EVERYTHING IN THE MEMORY SYSTEM WILL BE FORGOTTEN!")
            self.entries = []

    @property
    def is_empty(self) -> bool:
        """Return true if empty."""
        return len(self.entries) == 0

    @property
    def is_full(self) -> bool:
        """Return true if full."""
        return len(self.entries) == self.capacity

    @property
    def is_frozen(self) -> bool:
        """Is frozen?"""
        return self._frozen

    @property
    def size(self) -> int:
        """Get the size (number of filled entries) of the memory system."""
        return len(self.entries)

    def freeze(self) -> None:
        """Freeze the memory so that nothing can be added / deleted."""
        self._frozen = True

    def unfreeze(self) -> None:
        """Unfreeze the memory so that something can be added / deleted."""
        self._frozen = False

    def forget_random(self) -> None:
        """Forget a memory in the memory system in a uniform distribution manner."""
        logging.warning("forgetting a random memory using a uniform distribution ...")
        mem = random.choice(self.entries)
        self.forget(mem)

    def increase_capacity(self, increase: int) -> None:
        """Increase the capacity.

        Args
        ----
        increase: the amount of entries to increase.

        """
        assert isinstance(increase, int) and (not self.is_frozen)
        logging.debug(f"Increasing the memory capacity by {increase} ...")
        self.capacity += increase
        logging.info(
            f"The memory capacity has been increased by {increase} and now it's "
            f"{self.capacity}!"
        )

    def decrease_capacity(self, decrease: int) -> None:
        """decrease the capacity.

        Args
        ----
        decrease: the amount of entries to decrease.

        """
        assert (
            isinstance(decrease, int)
            and (self.capacity - decrease >= 0)
            and (not self.is_frozen)
        )
        logging.debug(f"Decreasing the memory capacity by {decrease} ...")
        self.capacity -= decrease
        logging.info(
            f"The memory capacity has been decreased by {decrease} and now it's "
            f"{self.capacity}!"
        )


class EpisodicMemory(Memory):
    """Episodic memory class."""

    def __init__(self, capacity: int) -> None:
        """Init an episodic memory system.

        Args
        ----
        capacity: capacity of the memory system (i.e., number of entries)

        """
        super().__init__("episodic", capacity)

    def can_be_added(self, mem: dict) -> bool:
        """Checks if a memory can be added to the system or not.

        Args
        ----
        mem: An episodic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        Returns
        -------
        True or False

        """
        if (self.capacity <= 0) or (self._frozen) or (self.is_full):
            return False

        else:
            return True

    def add(self, mem: dict) -> None:
        """Append a memory to the episodic memory system.

        Args
        ----
        mem: An episodic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if self._frozen:
            error_msg = "The memory system is frozen!"
            logging.error(error_msg)
            raise ValueError(error_msg)

        logging.debug(f"Adding a new memory entry {mem} ...")
        self.entries.append(mem)
        logging.info(
            f"memory entry {mem} added. Now there are in total of "
            f"{len(self.entries)} memories!"
        )
        self.clean_old_memories()

        # sort ascending
        self.entries.sort(key=lambda x: x["timestamp"])

        assert self.size <= self.capacity

    def get_oldest_memory(self, entries: list = None) -> List:
        """Get the oldest memory in the episodic memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        Returns
        -------
        mem: the oldest memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["timestamp"])[0]
        mem = random.choice(
            [mem for mem in entries if mem_candidate["timestamp"] == mem["timestamp"]]
        )
        logging.info(f"{mem} is the oldest memory in the entries.")

        return mem

    def get_latest_memory(self, entries: list = None) -> dict:
        """Get the latest memory in the episodic memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        Returns
        -------
        mem: the latest memory in a dictionary format

            for episodic:
            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["timestamp"])[-1]
        mem = random.choice(
            [mem for mem in entries if mem_candidate["timestamp"] == mem["timestamp"]]
        )
        logging.info(f"{mem} is the oldest memory in the entries.")

        return mem

    def forget_oldest(self) -> None:
        """Forget the oldest entry in the memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        """
        logging.debug("forgetting the oldest memory (FIFO)...")

        mem = self.get_oldest_memory()
        self.forget(mem)

    def answer_random(self) -> Tuple[str, int]:
        """Answer the question with a uniform-randomly chosen memory.

        Returns
        -------
        pred: prediction (e.g., desk)
        timestamp

        """
        if self.is_empty:
            logging.warning("Memory is empty. I can't answer any questions!")
            pred = None
            timestamp = None

        else:
            mem = random.choice(self.entries)
            pred = mem["object_location"]
            timestamp = mem["timestamp"]

        logging.info(f"pred: {pred}, timestamp: {timestamp}")

        return pred, timestamp

    def answer_latest(self, question: dict) -> Tuple[str, int]:
        """Answer the question with the latest relevant memory.

        If object X was found at Y and then later on found Z, then this strategy answers
        Z, instead of Y.

        Args
        ----
        question: a dict (i.e., {"human": <HUMAN>, "object": <OBJECT>})

        Returns
        -------
        pred: prediction
        timestamp: timestamp

        """
        logging.debug("answering a question with the answer_latest policy ...")

        if self.is_empty:
            logging.warning("Memory is empty. I can't answer any questions!")
            pred = None
            timestamp = None

        else:
            duplicates = get_duplicate_dicts(
                {"human": question["human"], "object": question["object"]}, self.entries
            )

            if len(duplicates) == 0:
                logging.info("no relevant memories found.")
                pred = None
                timestamp = None

            else:
                logging.info(
                    f"{len(duplicates)} relevant memories were found in the entries!"
                )
                mem = self.get_latest_memory(duplicates)
                pred = mem["object_location"]
                timestamp = mem["timestamp"]

        logging.info(f"pred: {pred}, timestamp: {timestamp}")

        return pred, timestamp

    @staticmethod
    def ob2epi(ob: dict) -> dict:
        """Turn an observation into an episodic memory.

        At the moment, the observation format is the same as an episodic memory
        for simplification.

        Args
        ----
        ob: An observation in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>,
            "object_location": <OBJECT_LOCATION>, "current_time": <CURRENT_TIME>}

        Returns
        -------
        mem: An episodic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        logging.debug(f"Turning an observation {ob} into a episodic memory ...")

        mem = deepcopy(ob)
        mem["timestamp"] = mem.pop("current_time")

        logging.info(f"Observation {ob} is now a episodic memory {mem}")

        return mem

    def find_same_memory(self, mem) -> dict:
        """Find an episodic memory that's almost the same as the query memory.

        Args
        ----
        mem: An episodic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>,
             "object_location": <OBJECT_LOCATION>, "timestamp": <TIMESTAMP>}

        Returns
        -------
        an episodic memory if it exists. Otherwise return None.

        """
        for entry in self.entries:
            if (
                (entry["human"] == mem["human"])
                and (entry["object"] == mem["object"])
                and (entry["object_location"] == mem["object_location"])
            ):
                return entry

        return None

    def clean_old_memories(self) -> List:
        """Find if there are duplicate memories with different timestamps."""
        logging.debug("finding if duplicate memories exist ...")

        entries = deepcopy(self.entries)
        logging.debug(f"There are {len(entries)} episdoic memories before cleaning")
        for entry in entries:
            del entry["timestamp"]

        entries = [str(mem) for mem in entries]  # to make list hashable
        uniques = set(entries)

        locs_all = [
            list_duplicates_of(entries, unique_entry) for unique_entry in uniques
        ]
        locs_all.sort(key=len)
        entries_cleaned = []

        for locs in locs_all:
            mem = self.entries[locs[0]]
            mem["timestamp"] = max([self.entries[loc]["timestamp"] for loc in locs])
            entries_cleaned.append(mem)

        self.entries = entries_cleaned
        logging.debug(f"There are {len(self.entries)} episdoic memories after cleaning")


class ShortMemory(Memory):
    """Short-term memory class."""

    def __init__(self, capacity: int) -> None:
        super().__init__("short", capacity)

    def add(self, mem: dict) -> None:
        """Append a memory to the short memory system.

        mem: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if self._frozen:
            error_msg = "The memory system is frozen!"
            logging.error(error_msg)
            raise ValueError(error_msg)

        assert not self.is_full

        logging.debug(f"Adding a new memory entry {mem} ...")
        self.entries.append(mem)
        logging.info(
            f"memory entry {mem} added. Now there are in total of "
            f"{len(self.entries)} memories!"
        )
        # sort ascending
        self.entries.sort(key=lambda x: x["timestamp"])

        assert self.size <= self.capacity

    def get_oldest_memory(self, entries: list = None) -> List:
        """Get the oldest memory in the episodic memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        Returns
        -------
        mem: the oldest memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["timestamp"])[0]
        mem = random.choice(
            [mem for mem in entries if mem_candidate["timestamp"] == mem["timestamp"]]
        )
        logging.info(f"{mem} is the oldest memory in the entries.")

        return mem

    def get_latest_memory(self, entries: list = None) -> dict:
        """Get the latest memory in the episodic memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        Returns
        -------
        mem: the latest memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["timestamp"])[-1]
        mem = random.choice(
            [mem for mem in entries if mem_candidate["timestamp"] == mem["timestamp"]]
        )
        logging.info(f"{mem} is the oldest memory in the entries.")

        return mem

    def forget_oldest(self) -> None:
        """Forget the oldest entry in the memory system.

        At the moment, this is simply done by looking up the timestamps and comparing
        them.

        """
        logging.debug("forgetting the oldest memory (FIFO)...")

        mem = self.get_oldest_memory()
        self.forget(mem)

    def find_similar_memories(self, mem) -> None:
        """Find similar memories.

        mem: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        logging.debug("Searching for similar memories in the short memory system...")
        similar = []
        for entry in self.entries:
            if (entry["object"] == mem["object"]) and (
                entry["object_location"] == mem["object_location"]
            ):
                similar.append(entry)

        logging.info(f"{len(similar)} similar short memories found!")

        return similar

    @staticmethod
    def ob2short(ob: dict) -> dict:
        """Turn an observation into an short memory.

        At the moment, the observation format is almost the same as an episodic memory
        for simplification.

        Args
        ----
        ob: An observation in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>,
            "object_location": <OBJECT_LOCATION>, "current_time": <CURRENT_TIME>}

        Returns
        -------
        mem: An episodic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        logging.debug(f"Turning an observation {ob} into a short memory ...")

        mem = deepcopy(ob)
        mem["timestamp"] = mem.pop("current_time")

        logging.info(f"Observation {ob} is now a episodic memory {mem}")

        return mem

    @staticmethod
    def short2epi(short: dict) -> dict:
        """Turn a short memory into a episodic memory.

        Args
        ----
        short: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        Returns
        -------
        epi: An episodic memory in a dictionary format

            {"human": <HUMAN>,
             "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        epi = deepcopy(short)
        return epi

    @staticmethod
    def short2sem(short: dict) -> dict:
        """Turn a short memory into a episodic memory.

        Args
        ----
        short: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        Returns
        -------
        sem: A semantic memory in a dictionary format

            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        sem = deepcopy(short)

        del sem["human"]
        del sem["timestamp"]
        sem["num_generalized"] = 1

        return sem

    def find_same_memory(self, mem) -> dict:
        """Find a short memory that's almost the same as the query memory.

        Args
        ----
        mem: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>,
             "object_location": <OBJECT_LOCATION>, "timestamp": <TIMESTAMP>}

        Returns
        -------
        A short memory if it exists. Otherwise return None.

        """
        for entry in self.entries:
            if (
                (entry["human"] == mem["human"])
                and (entry["object"] == mem["object"])
                and (entry["object_location"] == mem["object_location"])
            ):
                return entry

        return None


class SemanticMemory(Memory):
    """Semantic memory class."""

    def __init__(
        self,
        capacity: int,
    ) -> None:
        """Init a semantic memory system.

        Args
        ----
        capacity: capacity of the memory system (i.e., number of entries)

        """
        super().__init__("semantic", capacity)

    def can_be_added(self, mem: dict) -> bool:
        """Checks if a memory can be added to the system or not.

        Args
        ----
        True or False

        """
        if self.capacity <= 0:
            return False

        if self._frozen:
            return False

        if self.is_full:
            if self.find_same_memory(mem) is None:
                return False
            else:
                return True
        else:
            return True

    def pretrain_semantic(
        self,
        semantic_knowledge: dict,
        return_remaining_space: bool = True,
        freeze: bool = True,
    ) -> int:
        """Pretrain the semantic memory system from ConceptNet.

        Args
        ----
        semantic_knowledge: from ConceptNet.
        return_remaining_space: whether or not to return the remaining space from the
            semantic memory system.
        freeze: whether or not to freeze the semantic memory system or not.

        Returns
        -------
        free_space: free space that was not used, if any, so that it can be added to
            the episodic memory system.
        """
        self.semantic_knowledge = deepcopy(semantic_knowledge)
        for obj, loc in self.semantic_knowledge.items():
            if self.is_full:
                break
            mem = {"object": obj, "object_location": loc, "num_generalized": 1}
            logging.debug(f"adding a pretrained semantic knowledge {mem}")
            self.add(mem)

        if return_remaining_space:
            free_space = self.capacity - len(self.entries)
            self.decrease_capacity(free_space)
            logging.info(
                f"The remaining space {free_space} will be returned. Now "
                f"the capacity of the semantic memory system is {self.capacity}"
            )

        else:
            free_space = None

        if freeze:
            self.freeze()
            logging.info("The semantic memory system is frozen!")

        return free_space

    def get_weakest_memory(self, entries: list = None) -> List:
        """Get the weakest memory in the semantic memory system system.

        At the moment, this is simply done by looking up the number of generalized
        memories comparing them. In the end, an RL agent has to learn this
        by itself.

        Returns
        -------
        mem: the weakest memory in a dictionary format

            for semantic:
            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["num_generalized"])[0]
        mem = random.choice(
            [
                mem
                for mem in entries
                if mem_candidate["num_generalized"] == mem["num_generalized"]
            ]
        )
        logging.info(f"{mem} is the weakest memory in the entries.")

        return mem

    def get_strongest_memory(self, entries: list = None) -> List:
        """Get the strongest memory in the semantic memory system system.

        At the moment, this is simply done by looking up the number of generalized
        memories comparing them.

        Returns
        -------
        mem: the strongest memory in a dictionary format


            for semantic:
            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        if entries is None:
            logging.debug("No entries were specified. We'll use the memory system.")
            entries = self.entries

        # sorted() is ascending by default.
        mem_candidate = sorted(entries, key=lambda x: x["num_generalized"])[-1]
        mem = random.choice(
            [
                mem
                for mem in entries
                if mem_candidate["num_generalized"] == mem["num_generalized"]
            ]
        )
        logging.info(f"{mem} is the strongest memory in the entries.")

        return mem

    def find_similar_memories(self, mem) -> None:
        """Find similar memories.

        mem: A short memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "timestamp": <TIMESTAMP>}

        """
        logging.debug("Searching for similar memories in the short memory system...")
        similar = []
        for entry in self.entries:
            if (entry["object"] == mem["object"]) and (
                entry["object_location"] == mem["object_location"]
            ):
                similar.append(entry)

        logging.info(f"{len(similar)} similar short memories found!")

        return similar

    def forget_weakest(self) -> None:
        """Forget the weakest entry in the semantic memory system.

        At the moment, this is simply done by looking up the number of generalized
        memories and comparing them.

        """
        logging.debug("forgetting the weakest memory ...")
        mem = self.get_weakest_memory()
        self.forget(mem)
        logging.info(f"{mem} is forgotten!")

    def answer_random(self) -> Tuple[str, int]:
        """Answer the question with a uniform-randomly chosen memory.

        Returns
        -------
        pred: prediction (e.g., desk)
        num_generalized

        """
        if self.is_empty:
            logging.warning("Memory is empty. I can't answer any questions!")
            pred = None
            num_generalized = None

        else:
            mem = random.choice(self.entries)
            pred = mem["object_location"]
            num_generalized = mem["num_generalized"]

        logging.info(f"pred: {pred}, num_generalized: {num_generalized}")

        return pred, num_generalized

    def answer_strongest(self, question: list) -> Tuple[str, int]:
        """Answer the question (Find the head that matches the question, and choose the
        strongest one among them).

        Args
        ----
        question: a dict (i.e., {"human": <HUMAN>, "object": <OBJECT>})

        Returns
        -------
        pred: prediction
        num_generalized: number of generalized samples.

        """
        logging.debug("answering a question with the answer_strongest policy ...")

        if self.is_empty:
            logging.warning("Memory is empty. I can't answer any questions!")
            pred = None
            num_generalized = None

        else:
            duplicates = get_duplicate_dicts(
                {"object": question["object"]}, self.entries
            )
            if len(duplicates) == 0:
                logging.info("no relevant memories found.")
                pred = None
                num_generalized = None

            else:
                logging.info(
                    f"{len(duplicates)} relevant memories were found in the entries!"
                )
                mem = self.get_strongest_memory(duplicates)
                pred = mem["object_location"]
                num_generalized = mem["num_generalized"]

        logging.info(f"pred: {pred}, num_generalized: {num_generalized}")

        return pred, num_generalized

    @staticmethod
    def ob2sem(ob: dict) -> dict:
        """Turn an observation into a semantic memory.

        At the moment, this is simply done by removing the names from the head and the
        tail.

        Args
        ----
        ob: An observation in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>,
            "object_location": <OBJECT_LOCATION>, "current_time": <CURRENT_TIME>}


        Returns
        -------
        mem: A semantic memory in a dictionary format

            {"human": <HUMAN>, "object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        logging.debug(f"Turning an observation {ob} into a semantic memory ...")
        mem = deepcopy(ob)

        del mem["human"]
        del mem["current_time"]

        # 1 stands for the 1 generalized.
        mem["num_generalized"] = 1

        logging.info(f"Observation {ob} is now a semantic memory {mem}")

        return mem

    def clean_same_memories(self) -> List:
        """Find if there are duplicate memories cuz they should be summed out.

        At the moment, this is simply done by matching string values.

        """
        logging.debug("finding if duplicate memories exist ...")

        entries = deepcopy(self.entries)
        logging.debug(
            f"There are in total of {len(entries)} semantic memories before cleaning"
        )
        for entry in entries:
            del entry["num_generalized"]

        entries = [str(mem) for mem in entries]  # to make list hashable
        uniques = set(entries)

        locs_all = [
            list_duplicates_of(entries, unique_entry) for unique_entry in uniques
        ]
        locs_all.sort(key=len)
        entries_cleaned = []

        for locs in locs_all:
            mem = self.entries[locs[0]]
            mem["num_generalized"] = sum(
                [self.entries[loc]["num_generalized"] for loc in locs]
            )
            entries_cleaned.append(mem)

        self.entries = entries_cleaned
        logging.debug(
            f"There are now in total of {len(self.entries)} semantic memories after cleaning"
        )

    def add(self, mem: dict):
        """Append a memory to the semantic memory system.

        Args
        ----
        mem: A memory in a dictionary format

            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        """
        if self._frozen:
            error_msg = "The memory system is frozen!"
            logging.error(error_msg)
            raise ValueError(error_msg)

        logging.debug(f"Adding a new memory entry {mem} ...")
        self.entries.append(mem)
        logging.info(
            f"memory entry {mem} added. Now there are in total of "
            f"{len(self.entries)} memories!"
        )
        self.clean_same_memories()

        # sort ascending
        self.entries.sort(key=lambda x: x["num_generalized"])

        assert self.size <= self.capacity

    def find_same_memory(self, mem) -> dict:
        """Find a semantic memory that's almost the same as the query memory.

        Args
        ----
        mem: A semantic memory in a dictionary format

            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        Returns
        -------
        A semantic memory if it exists. Otherwise return None.

        """
        for entry in self.entries:
            if (entry["object"] == mem["object"]) and (
                entry["object_location"] == mem["object_location"]
            ):
                return entry

        return None

    def find_same_object_memory(self, mem) -> dict:
        """Find a semantic memory whose object is the same as the query memory.

        Args
        ----
        mem: A semantic memory in a dictionary format

            {"object": <OBJECT>, "object_location": <OBJECT_LOCATION>,
             "num_generalized": <NUM_GENERALIZED>}

        Returns
        -------
        A semantic memory if it exists. Otherwise return None.

        """
        for entry in self.entries:
            if (entry["object"] == mem["object"]) and (
                entry["object_location"] != mem["object_location"]
            ):
                return entry

        return None



"""Handcrafted / trained policies.

The trained neural network policies are not implemented yet.
"""
import random

from .memory import ShortMemory


def encode_observation(memory_systems: dict, policy: str, obs: dict) -> None:
    """Non RL policy of encoding an observation into a short-term memory.

    Args
    ----
    memory_systems: {"episodic": EpisodicMemory, "semantic": SemanticMemory,
                     "short": ShortMemory}
    policy: "argmax" or "neural"
    obs: observation = {"human": <human>,
                        "object": <obj>,
                        "object_location": <obj_loc>}

    """
    if policy.lower() == "argmax":
        mem_short = ShortMemory.ob2short(obs)
    else:
        raise NotImplementedError

    memory_systems["short"].add(mem_short)


def manage_memory(memory_systems: dict, policy: str) -> None:
    """Non RL memory management policy.

    Args
    ----
    memory_systems: {"episodic": EpisodicMemory, "semantic": SemanticMemory,
                     "short": ShortMemory}
    policy: "episodic", "semantic", "forget", "random", or "neural"

    """
    assert policy.lower() in [
        "episodic",
        "semantic",
        "forget",
        "random",
        "neural",
    ]
    if policy.lower() == "episodic":
        if memory_systems["episodic"].capacity != 0:
            if memory_systems["episodic"].is_full:
                memory_systems["episodic"].forget_oldest()
            mem_short = memory_systems["short"].get_oldest_memory()
            mem_epi = ShortMemory.short2epi(mem_short)
            memory_systems["episodic"].add(mem_epi)

    elif policy.lower() == "semantic":
        if memory_systems["semantic"].capacity != 0:
            if memory_systems["semantic"].is_full:
                memory_systems["semantic"].forget_weakest()
            mem_short = memory_systems["short"].get_oldest_memory()
            mem_sem = ShortMemory.short2sem(mem_short)
            memory_systems["semantic"].add(mem_sem)

    elif policy.lower() == "forget":
        pass

    elif policy.lower() == "random":
        action_number = random.choice([0, 1, 2])
        if action_number == 0:
            if memory_systems["episodic"].is_full:
                memory_systems["episodic"].forget_oldest()
            mem_short = memory_systems["short"].get_oldest_memory()
            mem_epi = ShortMemory.short2epi(mem_short)
            memory_systems["episodic"].add(mem_epi)

        elif action_number == 1:
            if memory_systems["semantic"].is_full:
                memory_systems["semantic"].forget_weakest()

            mem_short = memory_systems["short"].get_oldest_memory()
            mem_sem = ShortMemory.short2sem(mem_short)
            memory_systems["semantic"].add(mem_sem)

        else:
            pass

    elif policy.lower() == "neural":
        raise NotImplementedError

    else:
        raise ValueError

    memory_systems["short"].forget_oldest()


def answer_question(memory_systems: dict, policy: str, question: dict) -> str:
    """Non RL question answering policy.

    Args
    ----
    memory_systems: {"episodic": EpisodicMemory, "semantic": SemanticMemory,
                     "short": ShortMemory}
    policy: "episodic_semantic", "semantic_episodic", "episodic", "semantic",
            "random", or "neural",
    question: question = {"human": <human>, "object": <obj>}

    Returns
    -------
    pred: prediction

    """
    assert policy.lower() in [
        "episodic_semantic",
        "semantic_episodic",
        "episodic",
        "semantic",
        "random",
        "neural",
    ]
    pred_epi, _ = memory_systems["episodic"].answer_latest(question)
    pred_sem, _ = memory_systems["semantic"].answer_strongest(question)

    if policy.lower() == "episodic_semantic":
        if pred_epi is None:
            pred = pred_sem
        else:
            pred = pred_epi
    elif policy.lower() == "semantic_episodic":
        if pred_sem is None:
            pred = pred_epi
        else:
            pred = pred_sem
    elif policy.lower() == "episodic":
        pred = pred_epi
    elif policy.lower() == "semantic":
        pred = pred_sem
    elif policy.lower() == "random":
        pred = random.choice([pred_epi, pred_sem])
    elif policy.lower() == "neural":
        raise NotImplementedError
    else:
        raise ValueError

    return pred
